/* GSI - general sound interface   [-*- C++ -*-]
 *
 * load.c - shared sound server, file (sample) loading etc
 *   - sept 1996/may 1997/dec 1997/march-may 1998
 *
 * Copyright (C) 1996,1997,1998 W.H.Scholten
 *
 * Lower level API dependencies:
 *   - DSP mode setting: OSS but should not be too difficult to adapt
 *     to other systems.
 *   - OSS include locations are Linux/FreeBSD specific.
 *     Locations for other OS's?
 *   - sequencer/synth (MIDI's) playing: very OSS specific.
 */

/* This software is licensed under the CGPL version 0.4 or later
 * (see the file LICENSE.CGPL for details)
 */



/*---------------------------------------------------------------------------*
 *   .au/.snd file format:
 *----------------------------------------------
 *  ".snd"  START
 *  LENGTH  TYPE
 *  FREQ    CHANNELS
 *  <DESCRIPTION>    (text)
 *  <DATA>           (starts at START, length=LENGTH)
 *----------------------------------------------
 *   All are 32bit MSB-ints
 *   TYPE: 1 = mu-law compressed 8bit, (GSI_8BIT_MULAW)
 *         2 = uncompressed 8bit,      (GSI_8BIT)
 *         3 = 16bit uncompressed MSB  (GSI_16BIT)
 *   CHANNELS : 1 for mono, 2 for stereo.
 * internal: sound type: 0 =  8 bit mono, 1 = 16 bit mono, 2 =  8 bit stereo,
 *                       3 = 16 bit stereo
 *---------------------------------------------------------------------------*/


/*__________________INCLUDES___________________*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <errno.h>

#include <math.h>

#include <gsi/config.h>
#ifdef USE_OSS
#ifdef __linux__
#  include <sys/soundcard.h>
#endif
#ifdef __FreeBSD__
#  include <machine/soundcard.h>
#  include <sys/errno.h>
#endif
#endif
#ifdef USE_NEW_BSD
#  error code for new bsd driver not implemented
#endif
#ifdef USE_DEV_AUDIO
#  warning some includes for /dev/audio may be missing
#endif

#include <gsi/mytypes.h>
#include <gsi/debug.h>
#define _SNDSERVER_
#include <gsi/gsi_commands.h>
#include <gsi/gsi_server_version.h>
#include "formats.h"
#include "gsi_server.h"


/*__________________DEFINITIES___________________*/

/* Make these maxima dynamic? */
#define MAX_SOUNDS           256 /* Maximum number of sounds loaded (slots), this can have any value you want (note that it uses n*4 bytes for the 'playing' table), actually the interface sends 1 byte, so could use byte playing[..] and restrict to 256 always) */
/* perhaps use a 1-2 byte scheme to use 0-127, 128-255 +2nd byte for 128+32768 possible samples */

// These two should go to a separate config file:
#define MAX_STATIC_CHANNELS   64 /* Maximum number of simultaneous sounds on channels (the maximum of this value is 256) */
#define MAX_DYNAMIC_CHANNELS  32 /* Maximum number of simultaneous sounds dynamically allocated on a stack (the maximum of this value is only restricted by memory) */
#define MAX_CHANNELS        (MAX_STATIC_CHANNELS + MAX_DYNAMIC_CHANNELS)

#define MAX_STATUS_QUEUE_LEN 256

#define min(x,y) ((x)>(y) ? (y): (x))
#define max(x,y) ((x)<(y) ? (y): (x))

#define SIG(a,b,c,d) ( ((int)(a) <<24) | ((int)(b) <<16) | ((int)(c) << 8) | ((int)(d) ) )

/*__________________EXTERN___________________*/
extern char *server_name;

extern int  current_client;
extern struct client_data *current;

/*__________________VARIABELEN___________________*/

extern int  dsp_speed;
extern int  dsp_type;
extern int  dsp_log2_frag;
extern int  dsp_fragsize; /* samples=fragsize/channels/bytes */
extern int  dsp_channels;
extern int  dsp_bps; /* bytes per sample, 1 or 2 (determined from dsp_type) */

extern int  dsp_samples; // number of samples that fit in the fragsize


#if 0 /* all into the clients struct */
extern char sounds_dir[256];
extern int  num_sounds;       /* = 1+ highest number of sound slots used */


/* ** Sounds loaded into memory */
extern int  sounds_loaded;
/* have the sounds from the sound list been loaded? */
extern byte sound_from_list [MAX_SOUNDS]; /* incorporate into sound_flags ? */
/* is it a sound from the list or separately loaded? useful? */
extern char *sound_filename [MAX_SOUNDS]; /* file from which the slot was loaded */
/* RENAME: sound_filename [] */

extern int8 *sound_buffer [MAX_SOUNDS]; //the actual sound data
extern int  sound_size [MAX_SOUNDS];
/* bytes, >>should become samples<< ? (16/8 bit..) */
extern int  sound_speed [MAX_SOUNDS];
extern byte sound_type [MAX_SOUNDS];     /* 8/16bit, 8 bit mu-law */
extern byte sound_channels [MAX_SOUNDS]; /* mono, stereo */
extern int  sound_conversion [MAX_SOUNDS];
extern int  sound_conversion2 [MAX_SOUNDS]; /* recalculated at the moment */
/* conversion from sound to dsp speed (will be replaced by RL algorithm) */
#endif

/*__________________FUNCTIE PROTOTYPES___________________*/
void  snd_server();
extern int  setup_dsp(char *dspdev, int speed, int fragsize, int channels, int type);
void  play_fragment(int slot);

int   init_sounds_file(char *sndsfile_name);
void  read_soundsfile_sounds();
int   read_sound(int slot, char *file);
void  unload_sounds_file();
void  unload_sound(int slot);
extern void  init_sample(int slot, int speed, int type, int channels, int size);

extern void  set_sound_conversion(); /* all slots */
extern void  set_sound_conversion_of(int slot);

extern void  status_send_byte(int b);
extern void  status_flush();



int16 msb_int16(int fd); /* read MSB 16 bit int from file */
int32 msb_int32(int fd); /* read MSB 32 bit int from file */

int32 swap_int32(int a);
int16 swap_int16(int16 a);


/*__________________PROGRAMMA___________________*/

/*---------------------------------------------------------------------------*
 *  unload a sounds file and all the associated sounds.
 *---------------------------------------------------------------------------*/
/*static int sounds_file_loaded=False;*/
void unload_sounds_file() {
	int  i;

	if (current->sounds_file_loaded)
		for (i=0; i<MAX_SOUNDS; i++) {
			if (current->sound_from_list[i]) {
				//We only free sounds that were not loaded separately.
				unload_sound(i);
				free(current->sound_filename[i]);
				current->sound_from_list[i]=0;
			}
			current->sounds_file_loaded=False;
		}
}


/*---------------------------------------------------------------------------*
 *  unload a sound
 *---------------------------------------------------------------------------*/
void unload_sound(int slot) {
	DEB1(fprintf(stderr, "unload %d\n", slot));
	if (slot < /*max_slot*/ MAX_SOUNDS) {
		if (current->sound_buffer[slot]!=NULL) {
			free(current->sound_buffer[slot]);
			current->sound_buffer[slot]=NULL;
			current->sound_size[slot] = 0; /* used to test if sound was loaded in main loop */
		}
	}
	DEB1(fprintf(stderr, "done\n"));
}


/*---------------------------------------------------------------------------*
 *  Load the file listing all sounds used, and allocs memory for the complete
 *  paths. The last line is the number of sounds
 *---------------------------------------------------------------------------*/
int init_sounds_file(char *sndsfile_name) {
	int  i;
	char s[1024]; /* conversion string */
	char s2[1024];
	FILE *fp;


	//Unload old file.
	if (current->sounds_file_loaded) { unload_sounds_file(); current->sounds_file_loaded=False; }

	if (sndsfile_name==NULL) return -1;
	if (strlen(sndsfile_name)==0) return -1;

	strcpy(s, sndsfile_name);

	fp = fopen(s, "r");
	if( fp == 0 ) {
		fprintf (stderr, "%s: The sounds file %s could not be opened\n", server_name, sndsfile_name);
		status_send_byte(GSI_CALLBACK_NO_SUCH_FILE); /* should send slot no */
//status_flush();
		return -1;
	}
	for (i=0; i<256; i++) {
		fgets(s2, 127, fp); s2[127]=0;
		if ( feof(fp) ) break;
		if ( strlen(s2)==0 ) break; /* empty entry */
		s2[(int) strlen(s2) -1]=0; /* get rid of the linefeed */
		strcpy(s, current->sounds_dir);

		if ( s[(int)strlen(s) - 1] != '/' )
			strcat (s, "/");
		if ( s2[0] == '/' )
			strcpy (s, s2); /* absolute path */
		else
			strcat (s, s2); /* relative from the sounddir */

		if (current->sound_filename[i]!=NULL) {
			unload_sound(i);
			free(current->sound_filename[i]);
		}
		current->sound_filename[i] = (char *) malloc ((int)strlen (s)+1);

		strcpy (current->sound_filename[i], s);

		current->sound_buffer[i]=NULL;
		current->sound_size[i]=0;
	}
	fclose(fp);
	current->num_sounds=max(current->num_sounds, i); /* use previous 'num_sounds' */
	current->sounds_file_loaded=True;
	DEB1(fprt(stderr, "sounds file loaded\n"));
	return 0;
}


/*---------------------------------------------------------------------------*
 *  Read all sound files from the list from disk
 *---------------------------------------------------------------------------*/
void read_soundsfile_sounds() {
	/* load all sounds ( to prevent hickups when a sound is first used ) */
	int i;
	DEB1(fprt(stderr, "GSI_SERVER: read_soundsfile_sounds\n"));
	for (i=0; i<current->num_sounds; i++) {
		if (!read_sound(i, current->sound_filename[i]))
			current->sound_from_list[i]=1;
	}
	current->sounds_loaded=1;
}


/*---------------------------------------------------------------------------*
 *   Read a sound file from disk
 *   Slot -1 is used for a song (midi)?
 *---------------------------------------------------------------------------*/
int read_sound(int slot, char *sound_file) {
	int  i,fd,size, data_start;
	struct stat fd_stat;
	int  file_error=0;
	int  lsb_file=False;
	int  unsigned_linear_data=True;

	if ( slot>=current->num_sounds ) {
		fprintf (stderr, "%s: invalid sound number %d\n", server_name, slot);
		return(-1);
	}
	DEB1(fprt(stderr, "loading sound %d: %s\n",slot,sound_file));

	if (sound_file==NULL) {
		fprintf (stderr, "%s: undefined name for slot %d\n", server_name, slot);
		return -1;
	}

	fd = open(sound_file, O_RDONLY);
	if( fd <= 0 ) {
		fprintf (stderr, "%s: The sound %s could not be opened\n", server_name, sound_file);
		current->sound_size[slot]=-1;
		status_send_byte(GSI_CALLBACK_NO_SUCH_FILE); /* should send slot no */
//status_flush();
		return -1;
	}

	//Check that the file is a regular file (not a dir)
	fstat(fd, &fd_stat);
	if (S_ISREG(fd_stat.st_mode)) {
		int  name_len=strlen(sound_file);
		char *name_end=sound_file+name_len;
		int  file_format_recognized=False;

		/* Now find out what type of file we have*/
		/* raw files are recognized through the extension */
		if ( (name_len>4 && !strcmp(name_end-4,".raw")) ) {
			current->sound_speed[slot]=dsp_speed;
			current->sound_type[slot]=dsp_type;
			current->sound_channels[slot]=1; /* ? dsp_channels ? */
			size=lseek(fd,0,SEEK_END);
			data_start=0;
			file_format_recognized=True;
		}
		else { /* Check for file types by signature */
			int  file_sig;

			file_sig=msb_int32(fd);
			switch( file_sig) {
			case MAGIC_AU:
				size=lseek(fd,0,SEEK_END); /* see next */

				lseek(fd,0+4,SEEK_SET);
				data_start=msb_int32(fd);
//				size=msb_int32(fd);	/* BLASTED! In some files this is junk */
				size -=data_start;
				msb_int32(fd); /* ignore what? */

				current->sound_type[slot]=msb_int32(fd);
				current->sound_speed[slot]=msb_int32(fd);
				current->sound_channels[slot]=msb_int32(fd);
				file_format_recognized=True;
				DEB1(fprt(stderr, "size=%d, type=%d, speed=%d, channels=%d\n", size, current->sound_type[slot], current->sound_speed[slot], current->sound_channels[slot]));
				break; /* .au/.snd */

			case 0x52494646: /* "RIFF" = .wav, yuck! */
				msb_int32(fd);
				if (msb_int32(fd) != 0x57415645) break; /* "WAVE" */
				if (msb_int32(fd) != 0x666d7420) break; /* "fmt " */
				msb_int32(fd);
				if (msb_int16(fd) != 1*256) break; /* ?? can't remember */

				current->sound_channels[slot]=msb_int16(fd) >>8; /* Ha! */
				current->sound_speed[slot] = swap_int32(msb_int32(fd));
				lseek(fd,6,SEEK_CUR);
				current->sound_type[slot]=swap_int16(msb_int16(fd)) /8 +1;
				/* For 16 bit, do a lsb->msb swap for all pairs in the file */
				data_start=44;
				size=lseek(fd,0,SEEK_END) -data_start;
				lsb_file=True;
				file_format_recognized=True;
				break; /* .wav */

			case SIG('F', 'O', 'R', 'M'): /* AIFF/AIFC */
			{
				int  sig;
				int  tmp;
				int  chunk_len;
				int  found_data=False;

				size=msb_int32(fd); /* nog de andere headers eraf */
				sig=msb_int32(fd);
				if (sig != 'AIFF' && sig != 'AIFC') break;
				for (;;) {
					sig=msb_int32(fd);
					chunk_len=msb_int32(fd);
					DEB1(fprt(stderr, "chunk_len=%d\n", chunk_len));
					switch(sig) {
					case 'FVER':
					case 'INST': /* instrument chunk */
					case 'MARK':
						DEB1(fprt(stderr, "aiff/aifc: MARK/FVER/INST\n"));
						lseek(fd,chunk_len,SEEK_CUR);
						break;
					case 'COMM':
						current->sound_channels[slot]=msb_int16(fd);
						tmp=msb_int32(fd); /* sample frames */
						tmp=msb_int16(fd); /* bits per sample */
						DEB1(fprt(stderr, "aiff/aifc: COMM: ch=%d, bits=%d\n", current->sound_channels[slot], tmp));
						if (tmp==8) {
							current->sound_type[slot]=GSI_8BIT;
							unsigned_linear_data=False;
						}
						else if (tmp==16) current->sound_type[slot]=GSI_16BIT;

						{ /* Apple morons! using a long double, ridiculous */
							long double a;
							read(fd, &a, 10);
							/* Convert MSB to LSB for LSB machines */
#ifdef LITTLE_ENDIAN_MACHINE
							{
								char *aa=&a;
								int  z;
								for (z=0; z<5; z++) {
									int  tmp=aa[z]; aa[z]=aa[9-z]; aa[9-z]=tmp;
								}
							}
#endif
							DEB1(fprt(stderr, "aiff/aifc: COMM: long double=%f\n", (float) a));
							current->sound_speed[slot]=(int) a;
						}
						lseek(fd,chunk_len-10-8,SEEK_CUR);
						break;
					case 'SSND':
						tmp=msb_int32(fd); /* offset */
						tmp=msb_int32(fd); /* blocksize */
						size=chunk_len-8;
						data_start=lseek(fd,0,SEEK_CUR);
						found_data=True; /* should just seek till the end (can other blocks come after the SSND block?) */
						break;
					default:
						fprintf(stderr, "%s: unknown AIFF/C chunk type", server_name);
					}
					if (found_data) break;
				}
				file_format_recognized=True;
			}
			}
		}
		if (!file_format_recognized) {
			fprintf(stderr, "%s: unrecognized file format of %s\n", server_name, sound_file);
			return -1;
		}
		unload_sound(slot); /* can do this just as well earlier as the type/speed data may have been trashed */

		current->sound_buffer[slot]=(sbyte *) malloc(size);
		if(current->sound_buffer[slot]==NULL) {
			fprintf(stderr,"%s: couldn't allocate memory for sound\n", server_name);
			close(fd);
			return(-1);
		}
		current->sound_size[slot]=size;

		lseek(fd, data_start,SEEK_SET);
		read(fd, current->sound_buffer[slot],size);
	}
	else {
		DEB1(fprt(stderr, " slot %d not a regular file\n", slot));
		close(fd);
		return -1;
	}
	close(fd);

	/* Manipulate each sound type where needed */
	if (current->sound_type[slot]==GSI_8BIT && unsigned_linear_data) {
		/* unsigned 8 bit to signed 8 bit */
		for(i=0; i<size; i++) current->sound_buffer[slot][i] ^=0x80;
	}
	else if (current->sound_type[slot]==GSI_8BIT_MULAW) {
#if 0
		/* expansion to 16 bit */
		byte *tmpbuf;
		DEB1(fprintf(stderr, "GSI_8BIT_MULAW sample: expansion to 16 bit, size=%d\n", size));
		current->sound_type[slot]=GSI_16BIT;
		current->sound_size[slot]=size*2;
		tmpbuf=malloc(size*2);
		for (i=0; i<size; i++) {
			tmpbuf[i*2] = ((mulaw_to_linear16((int) (unsigned char) sound_buffer[slot][i])) >>8)&0xff;
			tmpbuf[i*2+1] = (mulaw_to_linear16((int) (unsigned char) sound_buffer[slot][i]))&0xff;
		}
		free(current->sound_buffer[slot]);
		current->sound_buffer[slot]=tmpbuf;
		size *=2;
#endif
	}
	else { /* 16 bit*/
		if (lsb_file) {
			for (i=0; i<size; i+=2) {
				byte sbi = current->sound_buffer[slot][i];
				current->sound_buffer[slot][i] = current->sound_buffer[slot][i+1];
				current->sound_buffer[slot][i+1] = sbi;
			}
		}
	}

	DEB1(fprt(stderr, " -> loaded, %d bytes\n",size));
	DEB1(fprt(stderr, "    speed: %d\n", current->sound_speed[slot]));
	//Conversion factors ..
	set_sound_conversion_of(slot);

	return(0);
}




/*---------------------------------------------------------------------------*
 *  Read a 32 or 16 bit msb-integer from a file
 *---------------------------------------------------------------------------*/
int32 msb_int32(int fd) {
	byte file_byte;
	int  i,int_return=0;

	for (i=3; i>=0; i--) {
		if (read(fd, &file_byte, 1) < 1) file_byte=0;
		int_return |= ((int)file_byte << (i*8));
	}
	return int_return;
}
int16 msb_int16(int fd) {
	byte file_byte;
	int  i,int_return=0;

	for (i=1; i>=0; i--) {
		if (read(fd, &file_byte, 1) < 1) file_byte=0;
		int_return |= ((int)file_byte << (i*8));
	}
	return int_return;
}
/* Convert endian-ness */
int32 swap_int32(int a) {
	return ((a & 255) << 24) | ((a& 0xff00) << 8) | ((a >> 8) & 0xff00) | ((a >>24) &0xff);
}
int16 swap_int16(int16 a) {
	return ((a & 0xff) << 8) | ((a >> 8) & 0xff);
}


/*---------------------------------------------------------------------------*
 *	EOF load.c
 *---------------------------------------------------------------------------*/
